Improve state-macro encapsulation.
[open-adventure.git] / advent.h
index 100de43cceb647e6b92bd2db9de80a9a414eb7d8..4188b1707d22616dfddb8e144ed1e3ae5184d29f 100644 (file)
--- a/advent.h
+++ b/advent.h
@@ -34,6 +34,7 @@
 #define BATTERYLIFE    2500       // turn limit increment from batteries
 #define WORD_NOT_FOUND -1         // "Word not found" flag value for the vocab hash functions.
 #define WORD_EMPTY     0          // "Word empty" flag value for the vocab hash functions
+#define PIT_KILL_PROB  35        // Percentage probability of dying from fall in pit.
 #define CARRIED        -1         // Player is toting it
 #define READ_MODE      "rb"       // b is not needed for POSIX but harmless
 #define WRITE_MODE     "wb"       // b is not needed for POSIX but harmless
 #define IS_FIXED -1
 #define IS_FREE 0
 
-/* Map a state property value to a negative range, where the object cannot be
- * picked up but the value can be recovered later.  Avoid colliding with -1,
- * which has its own meaning as STATE_NOTFOUND. */
-#define STASHED(obj)   (-1 - game.prop[obj])
+#ifndef FOUNDBOOL
+/* (ESR) It is fitting that translation of the original ADVENT should
+ * have left us a maze of twisty little conditionals that resists all
+ * understanding.  Setting and use of what is now the per-object state
+ * member (which used to be an array of its own) is our mystery. This
+ * state tangles together information about whether the object is a
+ * treasure, whether the player has seen it yet, and its activation
+ * state.
+ *
+ * Things we think we know:
+ *
+ * STATE_NOTFOUND is only set on treasures. Non-treasures start the
+ * game in STATE_FOUND.
+ *
+ * PROP_STASHED is supposed to map a state property value to a
+ * negative range, where the object cannot be picked up but the value
+ * can be recovered later.  Various objects get this peoperty when
+ * the cave starts to close. On;y seems to be signifucant for the bird 
+ * and readable objects, notably the clam/oyster - but the code around
+ * those test is difficult to read. */
+#define PROP_STASHIFY(n)       (-1 - (n))
+#define PROP_IS_STASHED(obj)   (game.objects[obj].prop < STATE_NOTFOUND)
+#define PROP_IS_NOTFOUND(obj)  (game.objects[obj].prop == STATE_NOTFOUND)
+#define PROP_IS_FOUND(obj)     (game.objects[obj].prop == STATE_FOUND)
+#define PROP_IS_STASHED_OR_UNSEEN(obj) (game.objects[obj].prop < 0)
+#define PROP_SET_FOUND(obj)    (game.objects[obj].prop = STATE_FOUND)
+#define PROP_SET_NOT_FOUND(obj)        (game.objects[obj].prop = STATE_NOTFOUND)
+#define PROP_IS_NOTFOUND2(g, o)        (g.objects[o].prop == STATE_NOTFOUND)
+#define PROP_IS_INVALID(val)   (val < -MAX_STATE - 1 || val > MAX_STATE)
+#else
+/* (ESR) Only the boldest of adventurers will explore here.  This
+ * alternate set of definitions for the macros above was an attempt to
+ * break from out of the state encoding a per-object "found" member
+ * telling whether or not the player has seen the object. 
+ *
+ * What's broken when you try to use thus is
+ * PROP_IS_STASHED_OR_UNSEEN. The symptom is game.tally getting
+ * decremented on non-treasures.
+ */
+#define PROP_STASHIFY(n)       (-(n))
+#define PROP_IS_STASHED(obj)   (game.objects[obj].prop < 0)
+#define PROP_IS_NOTFOUND(obj)  (!game.objects[obj].found)
+#define PROP_IS_FOUND(obj)     (game.objects[obj].found && game.objects[obj].prop == 0)
+#define PROP_IS_STASHED_OR_UNSEEN(obj) (!game.objects[obj].found || game.objects[obj].prop < 0)
+#define PROP_SET_FOUND(obj)    do {game.objects[obj].found = true; game.objects[obj].prop = STATE_FOUND;} while(0)
+#define PROP_SET_NOT_FOUND(obj)        game.objects[obj].found = false
+#define PROP_IS_NOTFOUND2(g, o)        (!g.objects[o].found)
+#define PROP_IS_INVALID(val)   (val < -MAX_STATE || val > MAX_STATE)
+#define PROP_SET_SEEN(obj)     game.objects[object].found = true
+#endif
+#define PROP_STASHED(obj)      PROP_STASHIFY(game.objects[obj].prop)
 
 #define PROMPT "> "
 
  */
 #define DESTROY(N)   move(N, LOC_NOWHERE)
 #define MOD(N,M)     ((N) % (M))
-#define TOTING(OBJ)  (game.place[OBJ] == CARRIED)
-#define AT(OBJ)      (game.place[OBJ] == game.loc || game.fixed[OBJ] == game.loc)
+#define TOTING(OBJ)  (game.objects[OBJ].place == CARRIED)
+#define AT(OBJ)      (game.objects[OBJ].place == game.loc || game.objects[OBJ].fixed == game.loc)
 #define HERE(OBJ)    (AT(OBJ) || TOTING(OBJ))
 #define CNDBIT(L,N)  (tstbit(conditions[L],N))
-#define LIQUID()     (game.prop[BOTTLE] == WATER_BOTTLE? WATER : game.prop[BOTTLE] == OIL_BOTTLE ? OIL : NO_OBJECT )
+#define LIQUID()     (game.objects[BOTTLE].prop == WATER_BOTTLE? WATER : game.objects[BOTTLE].prop == OIL_BOTTLE ? OIL : NO_OBJECT )
 #define LIQLOC(LOC)  (CNDBIT((LOC),COND_FLUID)? CNDBIT((LOC),COND_OILY) ? OIL : WATER : NO_OBJECT)
 #define FORCED(LOC)  CNDBIT(LOC, COND_FORCED)
-#define DARK(DUMMY)  (!CNDBIT(game.loc,COND_LIT) && (game.prop[LAMP] == LAMP_DARK || !HERE(LAMP)))
+#define DARK(DUMMY)  (!CNDBIT(game.loc,COND_LIT) && (game.objects[LAMP].prop == LAMP_DARK || !HERE(LAMP)))
 #define PCT(N)       (randrange(100) < (N))
 #define GSTONE(OBJ)  ((OBJ) == EMERALD || (OBJ) == RUBY || (OBJ) == AMBER || (OBJ) == SAPPH)
 #define FOREST(LOC)  CNDBIT(LOC, COND_FOREST)
@@ -115,7 +163,6 @@ typedef enum scorebonus {none, splatter, defeat, victory} score_t;
 
 /* Phase codes for action returns.
  * These were at one time FORTRAN line numbers.
- * The values don't matter, but perturb their order at your peril.
  */
 typedef enum {
     GO_TERMINATE,
@@ -167,7 +214,7 @@ struct game_t {
     int holdng;                  // number of objects being carried
     int igo;                     // # uses of "go" instead of a direction
     int iwest;                   // # times he's said "west" instead of "w"
-    int knfloc;                  // knife location; 0 if none, -1 after caveat
+    loc_t knfloc;                // knife location; LOC_NOWERE if none, -1 after caveat
     turn_t limit;                // lifetime of lamp
     loc_t loc;                   // where player is now
     loc_t newloc;                // where player is going
@@ -178,21 +225,36 @@ struct game_t {
     int saved;                   // point penalty for saves
     int tally;                   // count of treasures gained
     int thresh;                  // current threshold for endgame scoring tier
-    turn_t trndex;               // FIXME: not used, remove on next format bump
+    bool seenbigwords;           // have we red the graffiti in the Giant's Room? 
     turn_t trnluz;               // # points lost so far due to turns used
     turn_t turns;                // counts commands given (ignores yes/no)
     char zzword[TOKLEN + 1];     // randomly generated magic word from bird
-    int abbrev[NLOCATIONS + 1];  // has location been seen?
-    int atloc[NLOCATIONS + 1];   // head of object linked list per location
-    int dseen[NDWARVES + 1];     // true if dwarf has seen him
-    loc_t dloc[NDWARVES + 1];    // location of dwarves, initially hard-wired in
-    loc_t odloc[NDWARVES + 1];   // prior loc of each dwarf, initially garbage
-    loc_t fixed[NOBJECTS + 1];   // fixed location of object (if  not IS_FREE)
+    struct {
+       int abbrev;              // has location been seen?
+       int atloc;               // head of object linked list per location
+    } locs[NLOCATIONS + 1];
+    struct {
+       int seen;                // true if dwarf has seen him
+       loc_t loc;               // location of dwarves, initially hard-wired in
+       loc_t oldloc;            // prior loc of each dwarf, initially garbage
+    } dwarves[NDWARVES + 1];
+    struct {
+#ifdef FOUNDBOOL
+       bool found;              // has the location of this object bween found?
+#endif
+       loc_t fixed;             // fixed location of object (if not IS_FREE)
+        int prop;                // object state */
+       loc_t place;             // location of object
+    } objects[NOBJECTS + 1];
+    struct { 
+       bool used;               // hints[i].used = true iff hint i has been used.
+#ifndef FOUNDBOOL
+       int lc;                 // hints[i].lc = show int at LOC with cond bit i
+#else
+       int lc;                  // hints[i].lc = show int at LOC with cond bit i
+#endif
+    } hints[NHINTS];
     obj_t link[NOBJECTS * 2 + 1];// object-list links
-    loc_t place[NOBJECTS + 1];   // location of object
-    int hinted[NHINTS];          // hinted[i] = true iff hint i has been used.
-    int hintlc[NHINTS];          // hintlc[i] = how int at LOC with cond bit i
-    int prop[NOBJECTS + 1];      // object state array */
 };
 
 /*
@@ -226,7 +288,34 @@ typedef struct {
     command_state_t state;
 } command_t;
 
+/*
+ * Bump on save format change.
+ *
+ * Note: Verify that the tests run clean before bumping this, then rebuild the check
+ * files afterwards.  Otherwise you will get a spurious failure due to the old version
+ * having been generated into a check file.
+ */
+#define SAVE_VERSION   30
+
+/*
+ * Goes at start of file so saves can be identified by file(1) and the like.
+ */
+#define ADVENT_MAGIC   "open-adventure\n"
+
+/*
+ * If you change the first three members, the resume function may not properly
+ * reject saves from older versions. Later members can change, but bump the version
+ * when you do that.
+ */
+struct save_t {
+    char magic[sizeof(ADVENT_MAGIC)];
+    int32_t version;
+    int32_t canary;
+    struct game_t game;
+};
+
 extern struct game_t game;
+extern struct save_t save;
 extern struct settings_t settings;
 
 extern char *myreadline(const char *);
@@ -241,7 +330,7 @@ extern bool silent_yes_or_no(void);
 extern bool yes_or_no(const char*, const char*, const char*);
 extern void juggle(obj_t);
 extern void move(obj_t, loc_t);
-extern loc_t put(obj_t, loc_t, int);
+extern void put(obj_t, loc_t, int);
 extern void carry(obj_t, loc_t);
 extern void drop(obj_t, loc_t);
 extern int atdwrf(loc_t);
@@ -251,7 +340,7 @@ extern void set_seed(int32_t);
 extern int32_t randrange(int32_t);
 extern int score(enum termination);
 extern void terminate(enum termination) __attribute__((noreturn));
-extern int savefile(FILE *, int32_t);
+extern int savefile(FILE *);
 #if defined ADVENT_AUTOSAVE
 extern void autosave(void);
 #endif
@@ -262,8 +351,7 @@ extern int initialise(void);
 extern phase_codes_t action(command_t);
 extern void state_change(obj_t, int);
 extern bool is_valid(struct game_t);
-
-void bug(enum bugtype, const char *) __attribute__((__noreturn__));
+extern void bug(enum bugtype, const char *) __attribute__((__noreturn__));
 
 /* represent an empty command word */
 static const command_word_t empty_command_word = {